/*
  Copyright © 2001, 2022 Jan Bobrowski

  This program is free software; you can redistribute it or
  modify it under the terms of the GNU General Public License
  version 2 as published by the Free Software Foundation.
*/

#include "toico.h"

int is_gif(u8 s[SIGN_SIZE])
{
	return memcmp(s, "GIF8", 4) == 0 && (s[4] == '7' || s[4] == '9');
}

struct gif_reader {
	struct file *fi;
	struct image *im;
	u8 gpal[256][3]; int npal;
	int transp;

	int nblk; // remaining bytes in current img block

	int initcbits, cbits, bbits;
	u32 cbuf, cmask;

	int prefix[4096 + 2]; /* only >= FIRST used */
	u8 suffix[4096 + 2]; /* only >= FIRST used */
	int length[4096]; /* length - 1 */

	u8 stack[4096];
	int sp;

	u32 nxcode, clear;
};

static inline u32 mkcolor(u8 c[3])
{
	return c[0]<<16 | c[1]<<8 | c[2];
}

static void read_img(struct gif_reader *);

struct GIF_Header {
	u8 Signature[3], Version[3];
	u16le LogicalScreenWidth;
	u16le LogicalScreenHeight;
	u8 flags;
	u8 BackgroundColorIndex;
	u8 PixelAspectRatio;
};

enum {
	GIF_SizeOfGlobalColorTable = 0x07,
	GIF_H_Sort = 0x08,
	GIF_ColorResolution = 0x70,
	GIF_GlobalColorTable = 0x80
};

struct GIF_ImageDescriptor {
	/* u8 ImageSeparator = 0x2C */
	u16le ImageLeftPosition;
	u16le ImageTopPosition;
	u16le ImageWidth;
	u16le ImageHeight;
	u8 flags;
};

enum {
	GIF_SizeOfLocalColorTable = 0x07,
	GIF_I_Sort = 0x20,
	GIF_Interlace = 0x40,
	GIF_LocalColorTable = 0x80
};

struct GIF_GraphicControlExtension {
	u8 BlockSize;
	u8 flags;
	u16le DelayTime;
	u8 TransparentColorIndex;
};

void read_gif(struct image *img, struct file *fi)
{
	struct gif_reader gr;
	struct GIF_Header h;
	u32 background;
	int stop = 0;
	int v;

	gr.fi = fi;
	gr.im = img;

	read_all(gr.fi, &h, sizeof h);
	gr.im->width = get16le(h.LogicalScreenWidth);
	gr.im->height = get16le(h.LogicalScreenHeight);
	if (!gr.im->width || !gr.im->height)
		bad("Bad header");

	background = ALPHA;
	gr.npal = 0;
	if (h.flags & GIF_GlobalColorTable) {
		gr.npal = 2 << (h.flags & GIF_SizeOfGlobalColorTable);
		read_all(gr.fi, gr.gpal, 3*gr.npal);
#if 0
		if (h.bg < gr.npal)
			background = mkcolor(gr.gpal[h.bg]);
#endif
	}

	alloc_im(img);
	{
		u32 *p = img->data;
		u32 *e = data_end(img);
		do {
			*p++ = background;
		} while (p < e);
	}

	gr.transp = -1;
	for(;;) {
		int b = read_byte(gr.fi);
		if (b == EOF)
			break;
		if (b == ',') {
			read_img(&gr);
			if (stop)
				break;
			gr.transp = -1;
		} else if (b == '!') {
			v = read_byte(gr.fi);
			if (v == 0xF9) {
				struct GIF_GraphicControlExtension gce;
				read_all(gr.fi, &gce, sizeof gce);
				if (gce.BlockSize != 4)
					bad("Wrong GCE length");
				v = gce.flags & 0x1E;
				if (v && v != 4 || get16le(gce.DelayTime))
					stop = 1;
				if (gce.flags & 1)
					gr.transp = gce.TransparentColorIndex;
			}
			for (;;) {
				v = read_byte(gr.fi);
				if (!v)
					break;
				skip(gr.fi, v);
			}
		} else
			break;
	}
}

int get_code(struct gif_reader *gr)
{
	if (gr->cbits > gr->bbits) {
		do {
			int b;
			if (gr->nblk <= 0) {
				if (gr->nblk < 0)
					return -1;
				b = read_byte(gr->fi);
				gr->nblk = -1;
				if (b <= 0)
					break;
				gr->nblk = b;
			}
			b = read_byte(gr->fi);
			gr->nblk--;
			if (b < 0)
				break;
			gr->cbuf |= b << gr->bbits;
			gr->bbits += 8;
		} while (gr->bbits <= 24);
		if (gr->cbits > gr->bbits)
			return -1;
	}
	{
		int c = gr->cbuf & gr->cmask;
		gr->cbuf >>= gr->cbits;
		gr->bbits -= gr->cbits;
		return c;
	}
}

#define CLEAR (gr->clear)
#define END (gr->clear + 1)
#define FIRST (gr->clear + 2)

static int decode_reset(struct gif_reader *gr)
{
	int c;
	gr->cbits = gr->initcbits;
	gr->cmask = (1 << gr->cbits) - 1;
	do {
		c = get_code(gr);
	} while (c == CLEAR);
	if (c >= FIRST)
		bad("Bad code after reset");
	gr->nxcode = FIRST;
	gr->prefix[gr->nxcode] = c;
	gr->length[gr->nxcode] = 1;
	return c;
}

static void decode_start(struct gif_reader *gr)
{
	u8 c = read_byte(gr->fi);
	if (c < 1 || c > 11)
		errx(1, "Bad initial code size: %u\n", c);
	gr->bbits = 0;
	gr->initcbits = c + 1;
	gr->clear = 1 << c;
	gr->nblk = 0;

	memset(gr->length, 0, sizeof(gr->length[0])*CLEAR);
	c = decode_reset(gr);
	gr->sp = 1;
	gr->stack[0] = c;
}

static void inc_cbits(struct gif_reader *gr)
{
	if (gr->cbits == 12)
		return;
	gr->cbits++;
	gr->cmask = gr->cmask<<1 | 1;
}

static u32 *put_pixel(struct gif_reader *gr, u32 *p, int i)
{
	u32 c = gr->im->pal[i];
	if (i != gr->transp)
		*p = c;
	return p + 1;
}

static int decode(struct gif_reader *gr, u32 *outp, int size)
{
	u32 *oute = outp + size;
	int c;

	if (gr->sp) {
from_stack:
		do {
			outp = put_pixel(gr, outp, gr->stack[--gr->sp]);
			if (outp == oute)
				goto end;
		} while (gr->sp);
	}

	for (;;) {
		if (gr->nxcode > gr->cmask)
			inc_cbits(gr);
		c = get_code(gr);
		if (c < 0)
			break;
		gr->suffix[gr->nxcode] = c;
		gr->prefix[gr->nxcode + 1] = c;
		gr->length[gr->nxcode + 1] = gr->length[c] + 1;
		if (c >= CLEAR) {
			if (c <= END) {
				if (c == END)
					break;
				c = decode_reset(gr);
				goto out;
			} else if (c > gr->nxcode)
				bad("Bad code");

			assert(gr->prefix[c] < c);
			if (outp + gr->length[c] >= oute)
				goto do_not_fit;
			outp += gr->length[c];
			{
				int px = gr->prefix[c];
				u32 *p = outp - 1;
				while (px >= CLEAR) {
					put_pixel(gr, p--, gr->suffix[px]);
					assert(gr->prefix[px] < px);
					px = gr->prefix[px];
				}
				gr->suffix[gr->nxcode] = px;
				put_pixel(gr, p, px);
			}
			c = gr->suffix[c];
		}
		if (gr->nxcode < 4096)
			gr->nxcode++;
out:
		outp = put_pixel(gr, outp, c);
		if (outp >= oute)
			break;
	}
end:
	return oute - outp;

do_not_fit:
	{
		int px = gr->prefix[c];
		int i;
		assert(gr->sp == 0);
		i = 1;
		while (px >= CLEAR) {
			gr->stack[i++] = gr->suffix[px];
			assert(gr->prefix[px] < px);
			px = gr->prefix[px];
		}
		gr->stack[i] = px;
		gr->sp = i + 1;
		gr->suffix[gr->nxcode] = px;
		gr->stack[0] = gr->suffix[c];
		if (gr->nxcode < 4096)
			gr->nxcode++;
		goto from_stack;
	}
}

#define IMGAT(I,X,Y) pixel_at((I),(X),(Y))

static void read_img(struct gif_reader *gr)
{
	int ix, iy, iw, ih;
	u8 ipal[256][3];
	u8 (*palette)[3];
	struct GIF_ImageDescriptor h;
	int n, i;

	read_all(gr->fi, &h, sizeof h);

	ix = get16le(h.ImageLeftPosition);
	iy = get16le(h.ImageTopPosition);
	iw = get16le(h.ImageWidth);
	ih = get16le(h.ImageHeight);
	if (ix + iw > gr->im->width || iy + ih > gr->im->height)
		bad("Subimage do not fit");
	if (!(iw | ih))
		bad("No pixels in subimage");

	n = gr->npal;
	palette = gr->gpal;
	if (h.flags & GIF_LocalColorTable) {
		n = 2 << (h.flags & GIF_SizeOfLocalColorTable);
		read_all(gr->fi, ipal, 3*n);
		palette = ipal;
	}
	for (i=0; i<n; i++)
		gr->im->pal[i] = mkcolor(palette[i]);

	decode_start(gr);
	{
		int pass = 0;
		int step = gr->im->width;
		u32 *start = IMGAT(gr->im, ix, iy);
		u32 *end = IMGAT(gr->im, ix, iy + ih);
		u32 *p;
		if (h.flags & GIF_Interlace)
			step *= 8;
		for (p = start;;) {
			int r = decode(gr, p, iw);
			if (r)
				break;
			p += step;
			if (p < end)
				continue;

			if (step <= 2*gr->im->width)
				break;
			step = 4*gr->im->width >> pass++;
			p = start + step;
			step *= 2;
		}
	}

	if (gr->nblk < 0)
		return;

	/* skip the rest */
	if (gr->nblk)
		skip(gr->fi, gr->nblk);
	for (;;) {
		n = read_byte(gr->fi);
		if (n <= 0)
			break;
		skip(gr->fi, n);
	}
}
