/*
  Copyright © 2001, 2021, 2022 Jan Bobrowski

  This program is free software; you can redistribute it or
  modify it under the terms of the GNU General Public License
  version 2 as published by the Free Software Foundation.
*/

#include "toico.h"

int is_xpm(u8 p[SIGN_SIZE])
{
	return memcmp(p, "/* XPM *", 8) == 0 || memcmp(p, "! XPM2", 6) == 0;
}

enum mode {
	Unknown,
	XPM2 = 2,
	XPM3 = 3
};

enum state {
	Before,
	String,
	After
};

struct xpm_reader {
	struct file *fi;
	enum mode mode;
	enum state state;
	int lineno;
};

#define NOCOLOR 0x888888

static int get_char(struct xpm_reader *xr)
{
	int c;
	if (xr->state == After)
		return 0;

	if (xr->mode == Unknown) {
		c = read_byte(xr->fi);
		xr->mode = XPM3;
		if (c == '/')
			goto skip_comment;
		xr->mode = XPM2;
		do {
			c = read_byte(xr->fi);
			if (c == EOF)
				return 0;
		} while (c != '\n');
		xr->lineno++;
	}

	for (;;) {
		c = read_byte(xr->fi);
		if (c == EOF)
			return 0;
		if (c == 0)
			continue;
		if (c == '\n')
			xr->lineno++;
		if (xr->mode == XPM2) {
			if (c == '\n') {
				xr->state = After;
				c = 0;
			}
			return c;
		}
		if (xr->state == String) {
			if (c == '"') {
				xr->state = After;
				c = 0;
			}
			return c;
		}
		if (c == '"') {
			xr->state = String;
			continue;
		}
		if (c == '/') {
			int c0;
skip_comment:
			c = read_byte(xr->fi);
			if (c == EOF)
				return 0;
			if (c != '*') {
				if (c == '"')
					xr->state = String;
				continue;
			}
			c0 = 0;
			for (;;) {
				c = read_byte(xr->fi);
				if (c == EOF)
					return 0;
				if (c == '\n')
					xr->lineno++;
				if (c == '/' && c0 == '*')
					break;
				c0 = c;
			}
		}
	}
	return 0;
}

static void next(struct xpm_reader *xr)
{
	if (xr->state != After)
		while(get_char(xr));
	xr->state = Before;
}

static int skip_white(struct xpm_reader *xr)
{
	char c;
	while ((c = get_char(xr)))
		if (c > ' ')
			break;
	return c;
}

static u32 get_chars(struct xpm_reader *xr, int n)
{
	u32 v = 0;
	assert(n <= sizeof v);
	do {
		int c = get_char(xr);
		if (!c)
			return 0;
		v = v<<8 | c;
	} while (--n);
	return v;
}

static int hdigit(char c)
{
	return c <= '9' ? c - '0' : c - 'A' + 10;
}

static int hex2(char *t)
{
	return hdigit(t[0]) << 4 | hdigit(t[1]);
}

static int get_int(struct xpm_reader *xr)
{
	int v = 0;
	char c = skip_white(xr);
	do {
		if (c < '0' || c > '9')
			return -1;
		v = v*10 + c - '0';
		c = get_char(xr);
	} while (c > ' ');
	return v;
}

static u32 color_by_name(char*, int);

static u32 get_color(struct xpm_reader *xr, char *cp)
{
	char c, *p, *e;
	char name[32];
	int l, v;

	e = name;
	c = *cp;
	do {
		if (c >= 'a' && c <= 'z')
			c += 'A' - 'a';
		*e = c;
		if (e < endof(name) - 1)
			e++;
		c = get_char(xr);
	} while (c > ' ');
	*cp = c;
	*e = 0;

	if (name[0] != '#')
		return color_by_name(name, e - name);
	p = name + 1;
	l = e - name;
	if (l < 3)
		return NOCOLOR;
	if (l < 6) {
		v = hdigit(p[0]) << 16;
		v |= hdigit(p[1]) << 8;
		v |= hdigit(p[2]);
		v |= v << 4;
	} else {
		l /= 3;
		v = hex2(p) << 16; p += l;
		v |= hex2(p) << 8; p += l;
		v |= hex2(p);
	}
	return v;
}

struct cmap {
	unsigned key;
	u32 color;
};

void read_xpm(struct image *img, struct file *fi)
{
	struct xpm_reader xr;
	int i, x, y, nc, cpp;
	struct cmap *cmap;
	u32 *q;

	xr.fi = fi;
	xr.state = Before;
	xr.lineno = 1;
	xr.mode = Unknown;

	img->width = get_int(&xr);
	if (img->width < 1)
		goto badxpm;
	img->height = get_int(&xr);
	if (img->height < 1)
		goto badxpm;
	nc = get_int(&xr);
	if (nc < 1)
		goto badxpm;
	cpp = get_int(&xr);
	if (cpp < 1 || cpp > 4)
		goto badxpm;
	next(&xr);

	cmap = alloc(nc * sizeof *cmap);

	for (i = 0; i < nc; i++) {
		static char tt[] = "cg\347m";
		int l = sizeof tt;
		cmap[i].key = get_chars(&xr, cpp);
		if (!cmap[i].key)
			goto badxpm;
		cmap[i].color = 0;
		for (;;) {
			char c, cc, *p;
			cc = skip_white(&xr);
			if (!cc)
				break;
			c = get_char(&xr);
			if (c > ' ')
				cc |= c << 5;
			c = skip_white(&xr);
			if (!c)
				goto badxpm;
			p = strchr(tt, cc);
			if (p && p - tt < l) {
				l = p - tt;
				cmap[i].color = get_color(&xr, &c);
				if (!l)
					break;
			} else {
				do {
					c = get_char(&xr);
				} while (c > ' ');
			}
		}
		next(&xr);
	}

	alloc_im(img);

	q = img->data;
	for (y = 0; y < img->height; y++) {
		for (x = 0; x < img->width; x++) {
			unsigned c = get_chars(&xr, cpp);
			if (!c)
				goto badxpm;
			i = 0;
			while (c != cmap[i].key)
				if (++i == nc)
					goto badxpm;
			*q++ = cmap[i].color;
		}
		if (get_chars(&xr, cpp))
			goto badxpm;
		next(&xr);
	}

	free(cmap);

	return;

badxpm:
	bad("%d: Error in XPM", xr.lineno);
}

static char *cdic[] = {
	"ALICE", "ALMOND", "ANTIQUE", "AQUA", "AZURE", "BEIGE", "BISQUE", "BLACK",
	"BLANCHED", "BLUE", "BLUSH", "BRICK", "BROWN", "BURLYWOOD", "CADET",
	"CHART", "CHIFFON", "CHOCOLATE", "CORAL", "CORN", "CYAN", "DARK", "DEEP",
	"DEW", "DIM", "DODGER", "DRAB", "ENROD", "FIRE", "FLORAL", "FLOWER",
	"FOREST", "GAINSBORO", "GHOST", "GOLD", "GRAY", "GREEN", "GREY", "HONEY",
	"HOT", "INDIAN", "IVORY", "KHAKI", "LAVENDER", "LAWN", "LEMON", "LIGHT",
	"LIME", "LINEN", "MAGENTA", "MARINE", "MAROON", "MEDIUM", "MIDNIGHT",
	"MINTCREAM", "MISTY", "MOCCASIN", "NAVAJO", "NAVY", "NONE", "OLDLACE",
	"OLIVE", "ORANGE", "ORCHID", "PALE", "PAPAYA", "PEACH", "PERU", "PINK",
	"PLUM", "POWDER", "PUFF", "PURPLE", "RED", "REUSE", "ROSE", "ROSY", "ROYAL",
	"SADDLE", "SALMON", "SANDY", "SEA", "SHELL", "SIENNA", "SILK", "SKY",
	"SLATE", "SMOKE", "SNOW", "SPRING", "STEEL", "TAN", "THISTLE", "TOMATO",
	"TURQUOISE", "VIOLET", "WHEAT", "WHIP", "WHITE", "YELLOW",
};

static u32 colors[][2] = {
	{0xE,0xF0FFFF}, {0xF,0xF5F5DC}, {0x10,0xFFE4C4}, {0x11,0x000000},
	{0x13,0x0000FF}, {0x16,0xA52A2A}, {0x17,0xDEB887}, {0x1B,0xD2691E},
	{0x1C,0xFF7F50}, {0x1E,0x00FFFF}, {0x2A,0xDCDCDC}, {0x2C,0xFFD700},
	{0x2D,0xBEBEBE}, {0x2E,0x00FF00}, {0x2F,0xBEBEBE}, {0x33,0xFFFFF0},
	{0x34,0xF0E68C}, {0x35,0xE6E6FA}, {0x3A,0xFAF0E6}, {0x3B,0xFF00FF},
	{0x3D,0xB03060}, {0x40,0xF5FFFA}, {0x42,0xFFE4B5}, {0x44,0x000080},
	{0x45,0xFF000000}, {0x46,0xFDF5E6}, {0x48,0xFFA500}, {0x49,0xDA70D6},
	{0x4D,0xCD853F}, {0x4E,0xFFC0CB}, {0x4F,0xDDA0DD}, {0x52,0xA020F0},
	{0x53,0xFF0000}, {0x59,0xFA8072}, {0x5D,0xA0522D}, {0x62,0xFFFAFA},
	{0x65,0xD2B48C}, {0x66,0xD8BFD8}, {0x67,0xFF6347}, {0x68,0x40E0D0},
	{0x69,0xEE82EE}, {0x6A,0xF5DEB3}, {0x6C,0xFFFFFF}, {0x6D,0xFFFF00},
	{0xA13,0xF0F8FF}, {0xC6C,0xFAEBD7}, {0xD3C,0x7FFFD4}, {0xE01,0xF0FFFF},
	{0xE02,0xE0EEEE}, {0xE03,0xC1CDCD}, {0xE04,0x838B8B},
	{0x1001,0xFFE4C4}, {0x1002,0xEED5B7}, {0x1003,0xCDB79E},
	{0x1004,0x8B7D6B}, {0x120B,0xFFEBCD}, {0x1301,0x0000FF},
	{0x1302,0x0000EE}, {0x1303,0x0000CD}, {0x1304,0x00008B},
	{0x1369,0x8A2BE2}, {0x1601,0xFF4040}, {0x1602,0xEE3B3B},
	{0x1603,0xCD3333}, {0x1604,0x8B2323}, {0x1701,0xFFD39B},
	{0x1702,0xEEC591}, {0x1703,0xCDAA7D}, {0x1704,0x8B7355},
	{0x1813,0x5F9EA0}, {0x1954,0x7FFF00}, {0x1B01,0xFF7F24},
	{0x1B02,0xEE7621}, {0x1B03,0xCD661D}, {0x1B04,0x8B4513},
	{0x1C01,0xFF7256}, {0x1C02,0xEE6A50}, {0x1C03,0xCD5B45},
	{0x1C04,0x8B3E2F}, {0x1D5E,0xFFF8DC}, {0x1E01,0x00FFFF},
	{0x1E02,0x00EEEE}, {0x1E03,0x00CDCD}, {0x1E04,0x008B8B},
	{0x1F13,0x00008B}, {0x1F1E,0x008B8B}, {0x1F2D,0xA9A9A9},
	{0x1F2E,0x006400}, {0x1F2F,0xA9A9A9}, {0x1F34,0xBDB76B},
	{0x1F3B,0x8B008B}, {0x1F48,0xFF8C00}, {0x1F49,0x9932CC},
	{0x1F53,0x8B0000}, {0x1F59,0xE9967A}, {0x1F68,0x00CED1},
	{0x1F69,0x9400D3}, {0x204E,0xFF1493}, {0x222D,0x696969},
	{0x222F,0x696969}, {0x2313,0x1E90FF}, {0x2615,0xB22222},
	{0x276C,0xFFFAF0}, {0x292E,0x228B22}, {0x2B6C,0xF8F8FF},
	{0x2C01,0xFFD700}, {0x2C02,0xEEC900}, {0x2C03,0xCDAD00},
	{0x2C04,0x8B7500}, {0x2C25,0xDAA520}, {0x2E01,0x00FF00},
	{0x2E02,0x00EE00}, {0x2E03,0x00CD00}, {0x2E04,0x008B00},
	{0x2E6D,0xADFF2F}, {0x3021,0xF0FFF0}, {0x314E,0xFF69B4},
	{0x3253,0xCD5C5C}, {0x3301,0xFFFFF0}, {0x3302,0xEEEEE0},
	{0x3303,0xCDCDC1}, {0x3304,0x8B8B83}, {0x3401,0xFFF68F},
	{0x3402,0xEEE685}, {0x3403,0xCDC673}, {0x3404,0x8B864E},
	{0x3514,0xFFF0F5}, {0x362E,0x7CFC00}, {0x371A,0xFFFACD},
	{0x3813,0xADD8E6}, {0x381C,0xF08080}, {0x381E,0xE0FFFF},
	{0x382D,0xD3D3D3}, {0x382E,0x90EE90}, {0x382F,0xD3D3D3},
	{0x384E,0xFFB6C1}, {0x3859,0xFFA07A}, {0x386D,0xFFFFE0},
	{0x392E,0x32CD32}, {0x3B01,0xFF00FF}, {0x3B02,0xEE00EE},
	{0x3B03,0xCD00CD}, {0x3B04,0x8B008B}, {0x3D01,0xFF34B3},
	{0x3D02,0xEE30A7}, {0x3D03,0xCD2990}, {0x3D04,0x8B1C62},
	{0x3E13,0x0000CD}, {0x3E49,0xBA55D3}, {0x3E52,0x9370DB},
	{0x3E68,0x48D1CC}, {0x3F13,0x191970}, {0x4155,0xFFE4E1},
	{0x436C,0xFFDEAD}, {0x4413,0x000080}, {0x4724,0x6B8E23},
	{0x4801,0xFFA500}, {0x4802,0xEE9A00}, {0x4803,0xCD8500},
	{0x4804,0x8B5A00}, {0x4853,0xFF4500}, {0x4901,0xFF83FA},
	{0x4902,0xEE7AE9}, {0x4903,0xCD69C9}, {0x4904,0x8B4789},
	{0x4A2E,0x98FB98}, {0x4A68,0xAFEEEE}, {0x4B6B,0xFFEFD5},
	{0x4C51,0xFFDAB9}, {0x4E01,0xFFB5C5}, {0x4E02,0xEEA9B8},
	{0x4E03,0xCD919E}, {0x4E04,0x8B636C}, {0x4F01,0xFFBBFF},
	{0x4F02,0xEEAEEE}, {0x4F03,0xCD96CD}, {0x4F04,0x8B668B},
	{0x5013,0xB0E0E6}, {0x5201,0x9B30FF}, {0x5202,0x912CEE},
	{0x5203,0x7D26CD}, {0x5204,0x551A8B}, {0x5301,0xFF0000},
	{0x5302,0xEE0000}, {0x5303,0xCD0000}, {0x5304,0x8B0000},
	{0x5616,0xBC8F8F}, {0x5713,0x4169E1}, {0x5816,0x8B4513},
	{0x5901,0xFF8C69}, {0x5902,0xEE8262}, {0x5903,0xCD7054},
	{0x5904,0x8B4C39}, {0x5A16,0xF4A460}, {0x5B2E,0x2E8B57},
	{0x5B5C,0xFFF5EE}, {0x5D01,0xFF8247}, {0x5D02,0xEE7942},
	{0x5D03,0xCD6839}, {0x5D04,0x8B4726}, {0x5F13,0x87CEEB},
	{0x6013,0x6A5ACD}, {0x602D,0x708090}, {0x602F,0x708090},
	{0x6201,0xFFFAFA}, {0x6202,0xEEE9E9}, {0x6203,0xCDC9C9},
	{0x6204,0x8B8989}, {0x632E,0x00FF7F}, {0x6413,0x4682B4},
	{0x6501,0xFFA54F}, {0x6502,0xEE9A49}, {0x6503,0xCD853F},
	{0x6504,0x8B5A2B}, {0x6601,0xFFE1FF}, {0x6602,0xEED2EE},
	{0x6603,0xCDB5CD}, {0x6604,0x8B7B8B}, {0x6701,0xFF6347},
	{0x6702,0xEE5C42}, {0x6703,0xCD4F39}, {0x6704,0x8B3626},
	{0x6801,0x00F5FF}, {0x6802,0x00E5EE}, {0x6803,0x00C5CD},
	{0x6804,0x00868B}, {0x6953,0xD02090}, {0x6A01,0xFFE7BA},
	{0x6A02,0xEED8AE}, {0x6A03,0xCDBA96}, {0x6A04,0x8B7E66},
	{0x6C61,0xF5F5F5}, {0x6D01,0xFFFF00}, {0x6D02,0xEEEE00},
	{0x6D03,0xCDCD00}, {0x6D04,0x8B8B00}, {0x6D2E,0x9ACD32},
	{0xC6C01,0xFFEFDB}, {0xC6C02,0xEEDFCC}, {0xC6C03,0xCDC0B0},
	{0xC6C04,0x8B8378}, {0xD3C01,0x7FFFD4}, {0xD3C02,0x76EEC6},
	{0xD3C03,0x66CDAA}, {0xD3C04,0x458B74}, {0x181301,0x98F5FF},
	{0x181302,0x8EE5EE}, {0x181303,0x7AC5CD}, {0x181304,0x53868B},
	{0x195401,0x7FFF00}, {0x195402,0x76EE00}, {0x195403,0x66CD00},
	{0x195404,0x458B00}, {0x1D2813,0x6495ED}, {0x1D5E01,0xFFF8DC},
	{0x1D5E02,0xEEE8CD}, {0x1D5E03,0xCDC8B1}, {0x1D5E04,0x8B8878},
	{0x1F2C25,0xB8860B}, {0x1F472E,0x556B2F}, {0x1F4801,0xFF7F00},
	{0x1F4802,0xEE7600}, {0x1F4803,0xCD6600}, {0x1F4804,0x8B4500},
	{0x1F4901,0xBF3EFF}, {0x1F4902,0xB23AEE}, {0x1F4903,0x9A32CD},
	{0x1F4904,0x68228B}, {0x1F5B2E,0x8FBC8F}, {0x1F6013,0x483D8B},
	{0x1F602D,0x2F4F4F}, {0x1F602F,0x2F4F4F}, {0x204E01,0xFF1493},
	{0x204E02,0xEE1289}, {0x204E03,0xCD1076}, {0x204E04,0x8B0A50},
	{0x205F13,0x00BFFF}, {0x231301,0x1E90FF}, {0x231302,0x1C86EE},
	{0x231303,0x1874CD}, {0x231304,0x104E8B}, {0x261501,0xFF3030},
	{0x261502,0xEE2C2C}, {0x261503,0xCD2626}, {0x261504,0x8B1A1A},
	{0x2C2501,0xFFC125}, {0x2C2502,0xEEB422}, {0x2C2503,0xCD9B1D},
	{0x2C2504,0x8B6914}, {0x302101,0xF0FFF0}, {0x302102,0xE0EEE0},
	{0x302103,0xC1CDC1}, {0x302104,0x838B83}, {0x314E01,0xFF6EB4},
	{0x314E02,0xEE6AA7}, {0x314E03,0xCD6090}, {0x314E04,0x8B3A62},
	{0x325301,0xFF6A6A}, {0x325302,0xEE6363}, {0x325303,0xCD5555},
	{0x325304,0x8B3A3A}, {0x351401,0xFFF0F5}, {0x351402,0xEEE0E5},
	{0x351403,0xCDC1C5}, {0x351404,0x8B8386}, {0x371A01,0xFFFACD},
	{0x371A02,0xEEE9BF}, {0x371A03,0xCDC9A5}, {0x371A04,0x8B8970},
	{0x381301,0xBFEFFF}, {0x381302,0xB2DFEE}, {0x381303,0x9AC0CD},
	{0x381304,0x68838B}, {0x381E01,0xE0FFFF}, {0x381E02,0xD1EEEE},
	{0x381E03,0xB4CDCD}, {0x381E04,0x7A8B8B}, {0x382C25,0xEEDD82},
	{0x384E01,0xFFAEB9}, {0x384E02,0xEEA2AD}, {0x384E03,0xCD8C95},
	{0x384E04,0x8B5F65}, {0x385901,0xFFA07A}, {0x385902,0xEE9572},
	{0x385903,0xCD8162}, {0x385904,0x8B5742}, {0x385B2E,0x20B2AA},
	{0x385F13,0x87CEFA}, {0x386013,0x8470FF}, {0x38602D,0x778899},
	{0x38602F,0x778899}, {0x386413,0xB0C4DE}, {0x386D01,0xFFFFE0},
	{0x386D02,0xEEEED1}, {0x386D03,0xCDCDB4}, {0x386D04,0x8B8B7A},
	{0x3E0D3C,0x66CDAA}, {0x3E4901,0xE066FF}, {0x3E4902,0xD15FEE},
	{0x3E4903,0xB452CD}, {0x3E4904,0x7A378B}, {0x3E5201,0xAB82FF},
	{0x3E5202,0x9F79EE}, {0x3E5203,0x8968CD}, {0x3E5204,0x5D478B},
	{0x3E5B2E,0x3CB371}, {0x3E6013,0x7B68EE}, {0x3E632E,0x00FA9A},
	{0x3E6953,0xC71585}, {0x415501,0xFFE4E1}, {0x415502,0xEED5D2},
	{0x415503,0xCDB7B5}, {0x415504,0x8B7D7B}, {0x436C01,0xFFDEAD},
	{0x436C02,0xEECFA1}, {0x436C03,0xCDB38B}, {0x436C04,0x8B795E},
	{0x472401,0xC0FF3E}, {0x472402,0xB3EE3A}, {0x472403,0x9ACD32},
	{0x472404,0x698B22}, {0x485301,0xFF4500}, {0x485302,0xEE4000},
	{0x485303,0xCD3700}, {0x485304,0x8B2500}, {0x4A2C25,0xEEE8AA},
	{0x4A2E01,0x9AFF9A}, {0x4A2E02,0x90EE90}, {0x4A2E03,0x7CCD7C},
	{0x4A2E04,0x548B54}, {0x4A6801,0xBBFFFF}, {0x4A6802,0xAEEEEE},
	{0x4A6803,0x96CDCD}, {0x4A6804,0x668B8B}, {0x4A6953,0xDB7093},
	{0x4C5101,0xFFDAB9}, {0x4C5102,0xEECBAD}, {0x4C5103,0xCDAF95},
	{0x4C5104,0x8B7765}, {0x561601,0xFFC1C1}, {0x561602,0xEEB4B4},
	{0x561603,0xCD9B9B}, {0x561604,0x8B6969}, {0x571301,0x4876FF},
	{0x571302,0x436EEE}, {0x571303,0x3A5FCD}, {0x571304,0x27408B},
	{0x5B2E01,0x54FF9F}, {0x5B2E02,0x4EEE94}, {0x5B2E03,0x43CD80},
	{0x5B2E04,0x2E8B57}, {0x5B5C01,0xFFF5EE}, {0x5B5C02,0xEEE5DE},
	{0x5B5C03,0xCDC5BF}, {0x5B5C04,0x8B8682}, {0x5F1301,0x87CEFF},
	{0x5F1302,0x7EC0EE}, {0x5F1303,0x6CA6CD}, {0x5F1304,0x4A708B},
	{0x601301,0x836FFF}, {0x601302,0x7A67EE}, {0x601303,0x6959CD},
	{0x601304,0x473C8B}, {0x602D01,0xC6E2FF}, {0x602D02,0xB9D3EE},
	{0x602D03,0x9FB6CD}, {0x602D04,0x6C7B8B}, {0x632E01,0x00FF7F},
	{0x632E02,0x00EE76}, {0x632E03,0x00CD66}, {0x632E04,0x008B45},
	{0x641301,0x63B8FF}, {0x641302,0x5CACEE}, {0x641303,0x4F94CD},
	{0x641304,0x36648B}, {0x695301,0xFF3E96}, {0x695302,0xEE3A8C},
	{0x695303,0xCD3278}, {0x695304,0x8B2252}, {0x1F2C2501,0xFFB90F},
	{0x1F2C2502,0xEEAD0E}, {0x1F2C2503,0xCD950C}, {0x1F2C2504,0x8B6508},
	{0x1F472E01,0xCAFF70}, {0x1F472E02,0xBCEE68}, {0x1F472E03,0xA2CD5A},
	{0x1F472E04,0x6E8B3D}, {0x1F5B2E01,0xC1FFC1}, {0x1F5B2E02,0xB4EEB4},
	{0x1F5B2E03,0x9BCD9B}, {0x1F5B2E04,0x698B69}, {0x1F602D01,0x97FFFF},
	{0x1F602D02,0x8DEEEE}, {0x1F602D03,0x79CDCD}, {0x1F602D04,0x528B8B},
	{0x205F1301,0x00BFFF}, {0x205F1302,0x00B2EE}, {0x205F1303,0x009ACD},
	{0x205F1304,0x00688B}, {0x382C2501,0xFFEC8B}, {0x382C2502,0xEEDC82},
	{0x382C2503,0xCDBE70}, {0x382C2504,0x8B814C}, {0x382C256D,0xFAFAD2},
	{0x385F1301,0xB0E2FF}, {0x385F1302,0xA4D3EE}, {0x385F1303,0x8DB6CD},
	{0x385F1304,0x607B8B}, {0x38641301,0xCAE1FF}, {0x38641302,0xBCD2EE},
	{0x38641303,0xA2B5CD}, {0x38641304,0x6E7B8B}, {0x4A695301,0xFF82AB},
	{0x4A695302,0xEE799F}, {0x4A695303,0xCD6889}, {0x4A695304,0x8B475D},
};

static inline int find_chunk(char **p, char *e)
{
	int n, l = e - *p;
	for (n = 0; n < elemof(cdic); n++) {
		int cl = strlen(cdic[n]);
		if (cl > l)
			continue;
		if (memcmp(*p, cdic[n], cl) == 0) {
			*p += cl;
			return n;
		}
	}
	return -1;
}

static u32 color_by_name(char *t, int l)
{
	char *e = t + l;
	u32 n;

	if (l > 4 && l < 8
	   && (memcmp(t, "GRAY", 4) == 0 || memcmp(t, "GREY", 4) == 0)
	   && t[4] >= '0' && t[4] <= '9'
	) {
		n = (atoi(t + 4) * 255 + 49) / 100;
		return n | n<<8 | n<<16;
	}

	n = 0;
	while (t < e) {
		int v;
		if (*t > '0' && *t <= '9') {
			v = *t - '0';
			t++;
		} else {
			v = find_chunk(&t, e);
			if (v < 0)
				return NOCOLOR;
			v += 10;
		}
		n = n<<8 | v;
		if (n > 0x1000000) {
			if (*t)
				n = NOCOLOR;
			break;
		}
	}
	{
		int a = 0, b = elemof(colors);
		do {
			int c = a+b >> 1;
			if (colors[c][0] == n)
				return colors[c][1];
			if (colors[c][0] < n)
				a = c + 1;
			else
				b = c;
		} while (a < b);
	}
	return NOCOLOR;
}
